@@ -9,7 +9,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * SCCS: @(#) tclInterp.c 1.66 96/04/15 17:26:10
+ * SCCS: @(#) tclInterp.c 1.79 96/09/20 17:20:16
  */
 
 #include <stdio.h>
@@ -169,18 +169,18 @@ static char *TclCommandsToKeep[] = {
     "break",
     "case", "catch", "clock", "close", "concat", "continue",
     "eof", "error", "eval", "expr",
-    "fblocked", "fconfigure", "flush", "for", "foreach", "format",
+    "fblocked", "fileevent", "flush", "for", "foreach", "format",
     "gets", "global",
     "history",
     "if", "incr", "info", "interp",
     "join",
-    "lappend", "lindex", "linsert", "list", "llength", "lower", "lrange",
-    "lreplace", "lsearch", "lsort",
+    "lappend", "lindex", "linsert", "list", "llength",
+    "lower", "lrange", "lreplace", "lsearch", "lsort",
     "package", "pid", "proc", "puts",
     "read", "regexp", "regsub", "rename", "return",
-    "scan", "seek", "set", "split", "string", "switch",
-    "tell", "trace",
-    "unset", "update", "uplevel", "upvar",
+    "scan", "seek", "set", "split", "string", "subst", "switch",
+    "tell", "time", "trace",
+    "unset", "unsupported0", "update", "uplevel", "upvar",
     "vwait",
     "while",
     NULL};
@@ -344,6 +344,12 @@ MakeSafe(interp)
     /*
      * Remove the standard channels from the interpreter; safe interpreters
      * do not ordinarily have access to stdin, stdout and stderr.
+     *
+     * NOTE: These channels are not added to the interpreter by the
+     * Tcl_CreateInterp call, but may be added later, by another I/O
+     * operation. We want to ensure that the interpreter does not have
+     * these channels even if it is being made safe after being used for
+     * some time..
      */
 
     chan = Tcl_GetStdChannel(TCL_STDIN);
@@ -763,6 +769,8 @@ AliasHelper(curInterp, slaveInterp, masterInterp, masterPtr,
 {
     Alias *aliasPtr;			/* Storage for alias data. */
     Alias *tmpAliasPtr;			/* Temp storage for alias to delete. */
+    char *tmpAliasName;			/* Temp storage for name of alias
+                                         * to delete. */
     Tcl_HashEntry *hPtr;		/* Entry into interp hashtable. */
     int i;				/* Loop index. */
     int new;				/* Is it a new hash entry? */
@@ -844,12 +852,28 @@ AliasHelper(curInterp, slaveInterp, masterInterp, masterPtr,
         hPtr = Tcl_CreateHashEntry(&(slavePtr->aliasTable), aliasName, &new);
         if (new == 0) {
             tmpAliasPtr = (Alias *) Tcl_GetHashValue(hPtr);
-            (void) Tcl_DeleteCommand(slaveInterp, tmpAliasPtr->aliasName);
-            Tcl_DeleteHashEntry(hPtr);
+            tmpAliasName = Tcl_GetCommandName(slaveInterp,
+                    tmpAliasPtr->slaveCmd);
+            (void) Tcl_DeleteCommand(slaveInterp, tmpAliasName);
+
+            /*
+             * The hash entry should be deleted by the Tcl_DeleteCommand
+             * above, in its command deletion callback (most likely this
+             * will be AliasCmdDeleteProc, which does the deletion).
+             */
         }
     } while (new == 0);
     aliasPtr->aliasEntry = hPtr;
     Tcl_SetHashValue(hPtr, (ClientData) aliasPtr);
+    
+    /*
+     * Create the new command. We must do it after deleting any old command,
+     * because the alias may be pointing at a renamed alias, as in:
+     *
+     * interp alias {} foo {} bar		# Create an alias "foo"
+     * rename foo zop				# Now rename the alias
+     * interp alias {} foo {} zop		# Now recreate "foo"...
+     */
 
     targetPtr = (Target *) ckalloc((unsigned) sizeof(Target));
     targetPtr->slaveCmd = aliasPtr->slaveCmd;
@@ -973,7 +997,23 @@ DescribeAlias(interp, slaveInterp, aliasName)
     slavePtr = (Slave *) Tcl_GetAssocData(slaveInterp, "tclSlaveRecord",
             NULL);
     if (slavePtr == (Slave *) NULL) {
-        panic("DescribeAlias: could not find slave record");
+
+        /*
+         * It's possible that the interpreter still does not have a slave
+         * record. If so, create such a record now. This is only possible
+         * for interpreters that were created with Tcl_CreateInterp, not
+         * those created with Tcl_CreateSlave, so this interpreter does
+         * not have a master.
+         */
+        
+        slavePtr = (Slave *) ckalloc((unsigned) sizeof(Slave));
+        slavePtr->masterInterp = (Tcl_Interp *) NULL;
+        slavePtr->slaveEntry = (Tcl_HashEntry *) NULL;
+        slavePtr->slaveInterp = slaveInterp;
+        slavePtr->interpCmd = (Tcl_Command) NULL;
+        Tcl_InitHashTable(&(slavePtr->aliasTable), TCL_STRING_KEYS);
+        (void) Tcl_SetAssocData(slaveInterp, "tclSlaveRecord",
+                SlaveRecordDeleteProc, (ClientData) slavePtr);
     }
     hPtr = Tcl_FindHashEntry(&(slavePtr->aliasTable), aliasName);
     if (hPtr == (Tcl_HashEntry *) NULL) {
@@ -1017,7 +1057,9 @@ DeleteAlias(interp, slaveInterp, aliasName)
     slavePtr = (Slave *) Tcl_GetAssocData(slaveInterp, "tclSlaveRecord",
             NULL);
     if (slavePtr == (Slave *) NULL) {
-        panic("DeleteAlias: could not find slave record");
+        Tcl_AppendResult(interp, "alias \"", aliasName, "\" not found",
+                (char *) NULL);
+        return TCL_ERROR;
     }
     
     /*
@@ -1208,8 +1250,8 @@ Tcl_InterpCmd(clientData, interp, argc, argv)
     Tcl_Interp *masterInterp;		/* A master. */
     Master *masterPtr;			/* Master record for current interp. */
     Slave *slavePtr;			/* Record for slave interp. */
-    Tcl_HashEntry *hPtr;
-    Tcl_HashSearch hSearch;
+    Tcl_HashEntry *hPtr;		/* Search variable. */
+    Tcl_HashSearch hSearch;		/* Iteration variable. */
     size_t len;				/* Length of command name. */
     int result;				/* Result of eval. */
     char *cmdName;			/* Name of sub command to do. */
@@ -1275,24 +1317,6 @@ Tcl_InterpCmd(clientData, interp, argc, argv)
     }
 
     if (cmdName[0] == 'e') {
-        if ((strncmp(cmdName, "exists", len) == 0) && (len > 1)) {
-            if (argc > 3) {
-                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-                        " exists ?path?\"", (char *) NULL);
-                return TCL_ERROR;
-            }
-            if (argc == 3) {
-                if (GetInterp(interp, masterPtr, argv[2], NULL) ==
-                        (Tcl_Interp *) NULL) {
-                    Tcl_AppendResult(interp, "0", (char *) NULL);
-                } else {
-                    Tcl_AppendResult(interp, "1", (char *) NULL);
-                }
-            } else {
-                Tcl_AppendResult(interp, "1", (char *) NULL);
-            }
-            return TCL_OK;
-        }
         if ((strncmp(cmdName, "eval", len) == 0) && (len > 1)) {
             if (argc < 4) {
                 Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
@@ -1351,28 +1375,49 @@ Tcl_InterpCmd(clientData, interp, argc, argv)
             Tcl_Release((ClientData) slaveInterp);
             return result;        
         }
-    }
-
-    if ((cmdName[0] == 'i') && (strncmp(cmdName, "issafe", len) == 0)) {
-        if (argc > 3) {
-            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-                    " issafe ?path?\"", (char *) NULL);
-            return TCL_ERROR;
-        }
-        if (argc == 3) {
-            slaveInterp = GetInterp(interp, masterPtr, argv[2], &masterPtr);
-            if (slaveInterp == (Tcl_Interp *) NULL) {
-                Tcl_AppendResult(interp, "interpreter \"", argv[2],
-                        "\" not found", (char *) NULL);
+        if ((strncmp(cmdName, "exists", len) == 0) && (len > 2)) {
+            if (argc > 3) {
+                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
+                        " exists ?path?\"", (char *) NULL);
                 return TCL_ERROR;
             }
+            if (argc == 3) {
+                if (GetInterp(interp, masterPtr, argv[2], NULL) ==
+                        (Tcl_Interp *) NULL) {
+                    Tcl_AppendResult(interp, "0", (char *) NULL);
+                } else {
+                    Tcl_AppendResult(interp, "1", (char *) NULL);
+                }
+            } else {
+                Tcl_AppendResult(interp, "1", (char *) NULL);
+            }
+            return TCL_OK;
         }
-        if (masterPtr->isSafe == 0) {
-            Tcl_AppendResult(interp, "0", (char *) NULL);
-        } else {
-            Tcl_AppendResult(interp, "1", (char *) NULL);
+    }
+                
+    if (cmdName[0] == 'i') {
+        if ((len > 1) && (strncmp(cmdName, "issafe", len) == 0)) {
+            if (argc > 3) {
+                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
+                        " issafe ?path?\"", (char *) NULL);
+                return TCL_ERROR;
+            }
+            if (argc == 3) {
+                slaveInterp = GetInterp(interp, masterPtr, argv[2],
+                        &masterPtr);
+                if (slaveInterp == (Tcl_Interp *) NULL) {
+                    Tcl_AppendResult(interp, "interpreter \"", argv[2],
+                            "\" not found", (char *) NULL);
+                    return TCL_ERROR;
+                }
+            }
+            if (masterPtr->isSafe == 0) {
+                Tcl_AppendResult(interp, "0", (char *) NULL);
+            } else {
+                Tcl_AppendResult(interp, "1", (char *) NULL);
+            }
+            return TCL_OK;
         }
-        return TCL_OK;
     }
     
     if (cmdName[0] == 's') {
@@ -1589,81 +1634,85 @@ SlaveObjectCmd(clientData, interp, argc, argv)
         }
     }
     
-
-    if ((cmdName[0] == 'e') && (strncmp(cmdName, "eval", len) == 0)) {
-        if (argc < 3) {
-            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-                    " eval arg ?arg ...?\"", (char *) NULL);
-            return TCL_ERROR;
-        }
-
-        cmd = Tcl_Concat(argc-2, argv+2);
-        Tcl_Preserve((ClientData) slaveInterp);
-        result = Tcl_Eval(slaveInterp, cmd);
-        ckfree((char *) cmd);
-
-        /*
-         * Now make the result and any error information accessible. We have
-         * to be careful because the slave interpreter and the current
-         * interpreter can be the same - do not destroy the result.. This
-         * can happen if an interpreter contains an alias which is directed
-         * at a target command in the same interpreter.
-         */
-
-        if (interp != slaveInterp) {
-            if (result == TCL_ERROR) {
-
-               /*
-                * An error occurred, so transfer error information from the
-                * destination interpreter back to our interpreter.  Must clear
-                * interp's result before calling Tcl_AddErrorInfo, since
-                * Tcl_AddErrorInfo will store the interp's result in errorInfo
-                * before appending slaveInterp's $errorInfo;
-                * we've already got everything we need in the slave
-                * interpreter's $errorInfo.
-                */
-
-                Tcl_ResetResult(interp);
-                Tcl_AddErrorInfo(interp, Tcl_GetVar2(slaveInterp,
-                        "errorInfo", (char *) NULL, TCL_GLOBAL_ONLY));
-                Tcl_SetVar2(interp, "errorCode", (char *) NULL,
-                        Tcl_GetVar2(slaveInterp, "errorCode", (char *) NULL,
-                                TCL_GLOBAL_ONLY), TCL_GLOBAL_ONLY);
+    if (cmdName[0] == 'e') {
+        if ((len > 1) && (strncmp(cmdName, "eval", len) == 0)) {
+            if (argc < 3) {
+                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
+                        " eval arg ?arg ...?\"", (char *) NULL);
+                return TCL_ERROR;
             }
-            if (slaveInterp->freeProc != NULL) {
-                interp->result = slaveInterp->result;
-                interp->freeProc = slaveInterp->freeProc;
-                slaveInterp->freeProc = 0;
-            } else {
-                Tcl_SetResult(interp, slaveInterp->result, TCL_VOLATILE);
+
+            cmd = Tcl_Concat(argc-2, argv+2);
+            Tcl_Preserve((ClientData) slaveInterp);
+            result = Tcl_Eval(slaveInterp, cmd);
+            ckfree((char *) cmd);
+
+	    /*
+             * Make the result and any error information accessible. We have
+             * to be careful because the slave interpreter and the current
+             * interpreter can be the same - do not destroy the result.. This
+             * can happen if an interpreter contains an alias which is directed
+             * at a target command in the same interpreter.
+             */
+
+            if (interp != slaveInterp) {
+                if (result == TCL_ERROR) {
+
+                   /*
+                    * An error occurred, so transfer error information from the
+                    * destination interpreter back to our interpreter.  Clear
+                    * interp's result before calling Tcl_AddErrorInfo, since
+                    * Tcl_AddErrorInfo stores the interp's result in errorInfo
+                    * before appending slaveInterp's $errorInfo;
+                    * we've already got everything we need in the slave
+                    * interpreter's $errorInfo.
+                    */
+
+                    Tcl_ResetResult(interp);
+                    Tcl_AddErrorInfo(interp, Tcl_GetVar2(slaveInterp,
+                            "errorInfo", (char *) NULL, TCL_GLOBAL_ONLY));
+                    Tcl_SetVar2(interp, "errorCode", (char *) NULL,
+                            Tcl_GetVar2(slaveInterp, "errorCode",
+                                    (char *) NULL, TCL_GLOBAL_ONLY),
+                            TCL_GLOBAL_ONLY);
+                }
+                if (slaveInterp->freeProc != NULL) {
+                    interp->result = slaveInterp->result;
+                    interp->freeProc = slaveInterp->freeProc;
+                    slaveInterp->freeProc = 0;
+                } else {
+                    Tcl_SetResult(interp, slaveInterp->result, TCL_VOLATILE);
+                }
+                Tcl_ResetResult(slaveInterp);
             }
-            Tcl_ResetResult(slaveInterp);
+            Tcl_Release((ClientData) slaveInterp);
+            return result;
         }
-        Tcl_Release((ClientData) slaveInterp);
-        return result;
     }
 
-    if ((cmdName[0] == 'i') && (strncmp(cmdName, "issafe", len) == 0)) {
-        if (argc > 2) {
-            Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-                    " issafe\"", (char *) NULL);
-            return TCL_ERROR;
+    if (cmdName[0] == 'i') {
+        if ((len > 1) && (strncmp(cmdName, "issafe", len) == 0)) {
+            if (argc > 2) {
+                Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
+                        " issafe\"", (char *) NULL);
+                return TCL_ERROR;
+            }
+            masterPtr = (Master *) Tcl_GetAssocData(slaveInterp,
+                    "tclMasterRecord", NULL);
+            if (masterPtr == (Master *) NULL) {
+                panic("SlaveObjectCmd: could not find master record");
+            }
+            if (masterPtr->isSafe == 1) {
+                Tcl_AppendResult(interp, "1", (char *) NULL);
+            } else {
+                Tcl_AppendResult(interp, "0", (char *) NULL);
+            }
+            return TCL_OK;
         }
-        masterPtr = (Master *) Tcl_GetAssocData(slaveInterp,
-                "tclMasterRecord", NULL);
-        if (masterPtr == (Master *) NULL) {
-            panic("SlaveObjectCmd: could not find master record");
-        }
-        if (masterPtr->isSafe == 1) {
-            Tcl_AppendResult(interp, "1", (char *) NULL);
-        } else {
-            Tcl_AppendResult(interp, "0", (char *) NULL);
-        }
-        return TCL_OK;
     }
 
     Tcl_AppendResult(interp, "bad option \"", argv[1],
-            "\": should be alias, aliases, eval or issafe", (char *) NULL);
+            "\": should be alias, aliases, eval, or issafe", (char *) NULL);
     return TCL_ERROR;
 }
 
@@ -2080,9 +2129,10 @@ SlaveRecordDeleteProc(clientData, interp)
     }
         
     /*
-     * Finally dispose of the slave record itself.
+     * Finally dispose of the hash table and the slave record.
      */
-    
+
+    Tcl_DeleteHashTable(hTblPtr);
     ckfree((char *) slavePtr);    
 }
 
